<?php

namespace Slovocast\Domain\Repository;

use Slovocast\Infrastructure\DatabaseConnectionInterface;
use Slovocast\Infrastructure\User\UserAuthorizationInterface;
use Slovocast\Exception\EntityNotFoundException;
use Slovocast\Domain\Entity\User;
use PDO;

class UserRepository implements UserRepositoryInterface
{
    public function __construct(
        private DatabaseConnectionInterface $database,
        private UserAuthorizationInterface $userAuth
    ) {}

    /**
     * @param array $results The Query Results asking for all properties from
     *                       the database tables.
     * @return User
     */
    protected function userFromQueryResults(array $results): User
    {
        return User::fromArray([
            'id' => $results['id'],
            'email' => $results['email'],
            'password' => $results['password'],
            'name' => $results['name'],
            'createdAt' => $results['created_at'],
            'updatedAt' => $results['updated_at']
        ]);
    }

    /**
     * @param string $query The Query for getting a User
     * @param arary $params The parameters in the query
     * @param PDO $connection The PDO connection
     * @return array The column data from the Database
     * @throws EntityNotFoundException The User does no exist
     */
    protected function queryForUser(
        string $query,
        array $params,
        PDO $connection
    ): array {
        $statement = $connection->prepare($query);
        $statement->execute($params);
        $results = $statement->fetchAll();

        if (!$results) {
            throw new EntityNotFoundException("Unable to find User");
        }

        return array_shift($results);
    }

    /**
     * Get a single instance of the User Entity.
     *
     * @param int $id
     * @return User
     */
    public function get(int $id): User
    {
        $query = "SELECT * FROM users WHERE id = :id LIMIT 1";
        $userData = $this->queryForUser(
            $query,
            [ 'id' => $id ],
            $this->database->getConnection()
        );

        return $this->userFromQueryResults($userData);
    }

    /**
     * Get a single instance of the User Entity by their Email address.
     *
     * @param string $email
     * @return User
     */
    public function getFromEmail(string $email): User
    {
        $query = "SELECT * FROM users WHERE email = :email LIMIT 1";
        $userData = $this->queryForUser(
            $query, 
            [ 'email' => $email ], 
            $this->database->getConnection()
        );

        return $this->userFromQueryResults($userData);
    }

    public function save(User $user): bool
    {
        $connection = $this->database->getConnection();
        $query = "INSERT INTO users (email, password, name)
            VALUES (:email, :password, :name)";

        $statement = $connection->prepare($query);

        return $statement->execute([
            'email' => $user->getEmail(),
            'name'=> $user->getName(),
            'password' => $this->userAuth->hash($user->getPassword())
        ]);
    }

    public function update(User $user): bool
    {
        $connection = $this->database->getConnection();
        $query = "UPDATE users 
            SET email = :email,
                name = :name,
                password = :password
            WHERE id = :id";

        $statement = $connection->prepare($query);

        return $statement->execute([
            'email' => $user->getEmail(),
            'name' => $user->getName(),
            'password' => $this->userAuth->hash($user->getPassword()),
            'id' => $user->getId()
        ]);
    }

    /**
     * @param string $email
     * @param string $password
     * @return bool
     */
    public function verifyPassword(string $email, string $password): bool
    {
        try {
            $user = $this->getFromEmail($email);
            return $this->authUser->verify($password, $user->getPassword());
        } catch (\Exception $e) {
            return false;
        }
    }
}
